@@ -21,6 +21,22 @@
 
    ChangeLog
 
+   2004-01-22   Guillem Jover   <guillem@debian.org>
+
+   * Ported to 2.6 kernels.
+
+   2002/05/05   Guillem Jover   <guillem@hadrons.org>
+
+   * Re-included VOODOO pci id, accidentally removed in last patch,
+     thanks to Jan Kuipers to point this out.
+
+   2002/01/29   Guillem Jover   <guillem@hadrons.org>
+
+   * Fixed bug when compiling for 2.4 kernels with modversions.
+   * Added devfs support.
+   * Added MODULE_* declarations.
+   * Cleaned detection code.
+
    2000/02/04   Joseph Kain     <joseph@3dfx.com>
    
    * Updated Carlo Woord's email address.
@@ -100,16 +116,23 @@
 #endif
 
 #ifdef MODULE
-#include <linux/module.h>
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 #if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
 #define MODVERSIONS
 #endif
 #ifdef MODVERSIONS
 #include <linux/modversions.h>
 #endif
+#define mod_inc_use_count MOD_INC_USE_COUNT
+#define mod_dec_use_count MOD_DEC_USE_COUNT
 #else
-#define MOD_INC_USE_COUNT
-#define MOD_DEC_USE_COUNT
+#define mod_inc_use_count
+#define mod_dec_use_count
+#endif
+#include <linux/module.h>
+#else
+#define mod_inc_use_count
+#define mod_dec_use_count
 #endif
 
 #include <linux/types.h>
@@ -117,6 +140,9 @@
 #include <linux/mm.h>
 #include <linux/errno.h>
 #include <linux/pci.h>
+#if KERNEL_MIN_VER(2,3,46) || defined(DEVFS_SUPPORT)
+#include <linux/devfs_fs_kernel.h>
+#endif
 #include <asm/segment.h>
 #include <asm/ioctl.h>
 #include <asm/io.h>
@@ -137,6 +163,7 @@
 #define PCI_REVISION_ID_LINUX 0x8
 #define PCI_BASE_ADDRESS_0_LINUX 0x10
 #define PCI_BASE_ADDRESS_1_LINUX 0x14
+#define PCI_BASE_ADDRESS_2_LINUX 0x18
 #define SST1_PCI_SPECIAL1_LINUX 0x40
 #define SST1_PCI_SPECIAL2_LINUX 0x44
 #define SST1_PCI_SPECIAL3_LINUX 0x48
@@ -172,6 +199,22 @@
 #define PCI_DEVICE_ID_3DFX_VOODOO3 5
 #endif
 
+#ifndef PCI_DEVICE_ID_3DFX_VOODOO4
+#define PCI_DEVICE_ID_3DFX_VOODOO4 9
+#endif
+
+static struct pci_card {
+	unsigned short	vendor;
+	unsigned short	device;
+} pci_card_list[] = {
+	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO},
+	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO2},
+	{PCI_VENDOR_ID_ALLIANCE, 	PCI_DEVICE_ID_ALLIANCE_AT3D},
+	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_BANSHEE},
+	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO3},
+	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO4}
+};
+
 #ifdef DEBUG
 #define DEBUGMSG(x) printk x
 #else
@@ -186,6 +229,14 @@
 #define VM_OFFSET(vma) (vma->vm_pgoff << PAGE_SHIFT)
 #endif
 
+#if KERNEL_MIN_VER(2,6,0)
+#define my_remap_page_range(vma, start, ofs, len, prot) \
+        remap_page_range(vma, start, ofs, len, prot)
+#else
+#define my_remap_page_range(vma, start, ofs, len, prot) \
+        remap_page_range(start, ofs, len, prot)
+#endif
+
 struct pioData_t {
 	short port;
 	short size;
@@ -198,6 +249,7 @@ struct cardInfo_t {
 	int type;
 	int addr0;
   	int addr1;
+	int addr2;
 	unsigned char bus;
 	unsigned char dev;
 	struct file *curFile;
@@ -218,6 +270,9 @@ void cleanup_module(void);
 
 static cardInfo cards[MAXCARDS];
 static int numCards = 0;
+#if KERNEL_MIN_VER(2,3,46) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) || defined(DEVFS_SUPPORT)
+static devfs_handle_t devfs_handle;
+#endif
 
 static void findCardType(int vendor, int device)
 {
@@ -225,6 +280,7 @@ static void findCardType(int vendor, int device)
 	while (numCards < MAXCARDS && (dev = pci_find_device(vendor, device, dev))) {
 		pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &cards[numCards].addr0);
 		pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &cards[numCards].addr1);
+		pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &cards[numCards].addr2);
 		cards[numCards].bus = dev->bus->number;
 		cards[numCards].dev = dev->devfn;
 
@@ -243,14 +299,12 @@ static void findCardType(int vendor, int device)
 
 static int findCards(void)
 {
+	int i;
 	if (!pci_present())
 		return 0;
 	numCards = 0;
-	findCardType(PCI_VENDOR_ID_3DFX, PCI_DEVICE_ID_3DFX_VOODOO);
-	findCardType(PCI_VENDOR_ID_3DFX, PCI_DEVICE_ID_3DFX_VOODOO2);
-	findCardType(PCI_VENDOR_ID_ALLIANCE, 0x643d);
-	findCardType(PCI_VENDOR_ID_3DFX, PCI_DEVICE_ID_3DFX_BANSHEE);
-	findCardType(PCI_VENDOR_ID_3DFX, PCI_DEVICE_ID_3DFX_VOODOO3);
+	for (i = 0; i < (sizeof(pci_card_list)/sizeof(struct pci_card)); i++)
+		findCardType(pci_card_list[i].vendor, pci_card_list[i].device);
 	return numCards;
 }
 
@@ -261,7 +315,7 @@ static int open_3dfx(struct inode *inode, struct file *file)
 		printk("3dfx: No 3Dfx cards found\n");
 		return -ENODEV;
 	}
-	MOD_INC_USE_COUNT;
+	mod_inc_use_count;
 	return 0;
 }
 
@@ -273,7 +327,7 @@ static int release_3dfx(struct inode *inode, struct file *file)
 	for (i = 0; i < numCards; ++i)
 		if (cards[i].curFile == file)
 			cards[i].curFile = 0;
-	MOD_DEC_USE_COUNT;
+	mod_dec_use_count;
 
 	return 0;
 }
@@ -314,7 +368,7 @@ static int mmap_3dfx(struct file *file, struct vm_area_struct *vma)
 #if defined(__i386__)
 	pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
 #endif
-	if (remap_page_range(vma->vm_start, VM_OFFSET(vma), len, 
+	if (my_remap_page_range(vma, vma->vm_start, VM_OFFSET(vma), len, 
 			     vma->vm_page_prot)) {
 		DEBUGMSG(("3dfx: Page remap failed\n"));
 		return -EAGAIN;
@@ -359,6 +413,11 @@ static int doQueryFetch(pioData *desc)
 			return -EINVAL;
 		copy_to_user(desc->value, &cards[desc->device].addr1, desc->size);
 		return 0;
+	case PCI_BASE_ADDRESS_2_LINUX:
+		if (desc->size != 4)
+			return -EINVAL;
+		copy_to_user(desc->value, &cards[desc->device].addr2, desc->size);
+		return 0;
 	case SST1_PCI_SPECIAL1_LINUX:
 		if (desc->size != 4)
 			return -EINVAL;
@@ -475,10 +534,12 @@ static int doQuery(unsigned int cmd, unsigned long arg)
 static int doPIORead(pioData *desc)
 {
 	int retval;
-	char retchar;
+	int retchar;
 
 	if ((retval = verify_area(VERIFY_WRITE, desc->value, desc->size)))
 		return retval;
+#if 0
+	/* restricted */
 	switch (desc->port) {
 	case VGA_INPUT_STATUS_1C:
 		break;
@@ -491,12 +552,47 @@ static int doPIORead(pioData *desc)
 	default:
 		return -EPERM;
 	}
+#else
+	/* full range */
+	{
+         int i = desc->device;
+         unsigned short port = desc->port;
+         unsigned short base;
+         if (i < 0 || i >= numCards) {
+            /* scan for valid SSTIO aperture */
+            for (i = 0; i < numCards; i++) {
+                base = cards[i].addr2 & ~1;
+                if (base <= port && port <= (base + 0x107)) {
+                   break;
+                }
+            }
+            if (i == numCards) {
+               return -EPERM;
+            }
+         } else {
+            /* check the given SSTIO aperture */
+            base = cards[i].addr2 & ~1;
+            if (base > port || port > (base + 0x107)) {
+                return -EPERM;
+            }
+         }
+	}
+#endif
 
-	if (desc->size != 1)
+	switch (desc->size) {
+	case 1:
+		retchar = inb(desc->port);
+		break;
+	case 2:
+		retchar = inw(desc->port);
+		break;
+	case 4:
+		retchar = inl(desc->port);
+		break;
+	default:
 		return -EINVAL;
-
-	retchar = inb(desc->port);
-	copy_to_user(desc->value, &retchar, sizeof(char));
+	}
+	copy_to_user(desc->value, &retchar, desc->size);
 
 	return 0;
 }
@@ -504,11 +600,13 @@ static int doPIORead(pioData *desc)
 static int doPIOWrite(pioData *desc)
 {
 	int retval;
-	char retchar;
+	int retchar;
 
 	if ((retval = verify_area(VERIFY_READ, desc->value, desc->size)))
 		return retval;
 
+#if 0
+	/* restricted */
 	switch (desc->port) {
 	case SC_INDEX:
 		break;
@@ -519,12 +617,47 @@ static int doPIOWrite(pioData *desc)
 	default:
 		return -EPERM;
 	}
+#else
+	/* full range */
+	{
+         int i = desc->device;
+         unsigned short port = desc->port;
+         unsigned short base;
+         if (i < 0 || i >= numCards) {
+            /* scan for valid SSTIO aperture */
+            for (i = 0; i < numCards; i++) {
+                base = cards[i].addr2 & ~1;
+                if (base <= port && port <= (base + 0x107)) {
+                   break;
+                }
+            }
+            if (i == numCards) {
+               return -EPERM;
+            }
+         } else {
+            /* check the given SSTIO aperture */
+            base = cards[i].addr2 & ~1;
+            if (base > port || port > (base + 0x107)) {
+                return -EPERM;
+            }
+         }
+	}
+#endif
 
-	if (desc->size != 1)
+	copy_from_user(&retchar, desc->value, desc->size);
+	switch (desc->size) {
+	case 1:
+		outb(retchar, desc->port);
+		break;
+	case 2:
+		outw(retchar, desc->port);
+		break;
+	case 4:
+		outl(retchar, desc->port);
+		break;
+	default:
 		return -EINVAL;
-
-	copy_from_user(&retchar, desc->value, sizeof(char));
-	outb(retchar, desc->port);
+	}
 
 	return 0;
 }
@@ -673,6 +806,18 @@ int init_module(void)
 		printk("3dfx: Unable to register character device with major %d\n", MAJOR_3DFX);
 		return ret;
 	}
+
+#if KERNEL_MIN_VER(2,6,0)
+	devfs_mk_cdev(MKDEV(MAJOR_3DFX, DEVICE_VOODOO),
+		      S_IFCHR | S_IROTH | S_IWOTH | S_IRGRP | S_IWGRP,
+		      "3dfx");
+#elif KERNEL_MIN_VER(2,3,46) || defined(DEVFS_SUPPORT)
+	devfs_handle = devfs_register(NULL, "3dfx", DEVFS_FL_NONE,
+		MAJOR_3DFX, DEVICE_VOODOO,
+		S_IFCHR | S_IROTH | S_IWOTH | S_IRGRP | S_IWGRP,
+		&fops_3dfx, NULL);
+#endif
+
 	DEBUGMSG(("3dfx: Successfully registered device 3dfx\n"));
 	findCards();
 
@@ -698,6 +843,12 @@ void cleanup_module(void)
 #ifdef HAVE_MTRR
 	resetmtrr_3dfx();
 #endif
+#if KERNEL_MIN_VER(2,6,0)
+	devfs_remove("3dfx");
+#elif KERNEL_MIN_VER(2,3,46) || defined(DEVFS_SUPPORT)
+	devfs_unregister(devfs_handle);
+#endif
+
 	if (unregister_chrdev(MAJOR_3DFX, "3dfx"))
 	{
 	  DEBUGMSG(("3dfx: unregister_chrdev failed\n"));
@@ -719,6 +870,11 @@ long init_3dfx(long mem_start, long mem_end)
 }
 #endif /* !MODULE */
 
+#if KERNEL_MIN_VER(2,1,21)
+MODULE_AUTHOR("Daryll Strauss et al.");
+MODULE_DESCRIPTION("3dfx glide kernel device driver");
+MODULE_LICENSE("GPL");
+#endif
 
 #if defined(DEBUG)
 /*
